package be.iminds.ilabt.jfed.restapi.json;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import io.dropwizard.jackson.Jackson;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import java.net.URISyntaxException;
import java.util.*;

import static io.dropwizard.testing.FixtureHelpers.fixture;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.is;

@RunWith(Parameterized.class)
public class JsonTest {
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Parameterized.Parameters(name = "{index}: collection={0} index={1}")
    public static Collection<Object[]> data() throws Exception {
        List<Object[]> res = new ArrayList<>();

        ExperimentTestCollection sampleCollection = new ExperimentTestCollection();
        for (int i = 0; i < sampleCollection.getSize(); i++) {
            res.add(new Object[]{sampleCollection, i});
        }


        return res;
    }

    private final ExperimentTestCollection experimentTestCollection;
    private final int index;
    public JsonTest(ExperimentTestCollection experimentTestCollection, int index) throws URISyntaxException {
        if (index == 0) {
            (experimentTestCollection).setUri(0, "http://localhost:9000/api");
        }
        this.experimentTestCollection = experimentTestCollection;
        this.index = index;
    }

    @Test
    public void serializesToJSON() throws Exception {
        String fixtureFilename = experimentTestCollection.getFixtureByIndex(index);
        System.out.println("serializesToJSON is reading " + fixtureFilename);
        Experiment exp = experimentTestCollection.getByIndex(index);

        final String expected =  MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(
                MAPPER.readValue(
                        fixture(fixtureFilename),
                        exp.getClass())
        );

        final String actual = MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(exp);
        assertThat(actual, is(equalTo(expected)));
    }

    @Test
    public void serializesToJSONTree() throws Exception {
        String fixtureFilename = experimentTestCollection.getFixtureByIndex(index);
        System.out.println("serializesToJSON is reading " + fixtureFilename);
        Experiment exp = experimentTestCollection.getByIndex(index);

        final Map expected = MAPPER.readValue(fixture(fixtureFilename), Map.class);
        final Map actual = MAPPER.readValue(MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(exp), Map.class);

        assertThat(mapToString(actual), is(equalTo(mapToString(expected))));
        assertThat(mapToString(actual) + "\n\n != \n\n" + mapToString(expected), actual, is(equalTo(expected)));
    }

    @Test
    public void deserializesFromJSON() throws Exception {
        String fixtureFilename = experimentTestCollection.getFixtureByIndex(index);
        System.out.println("deserializesFromJSON is reading " + fixtureFilename);
        Experiment exp = experimentTestCollection.getByIndex(index);
        Experiment actualFedmonObject = MAPPER.readValue(fixture(fixtureFilename), exp.getClass());

//        assertThat(actualFedmonObject, is(equalTo(expectedFedmonObject)));

//        System.out.flush();
//        System.err.flush();
//        System.out.println("deserializesFromJSON is comparing actual="+actualFedmonObject+" expected="+expectedFedmonObject);
//        System.out.flush();
//        System.err.flush();
        experimentTestCollection.assertSame(actualFedmonObject, exp, index == 0);
        System.out.flush();
        System.err.flush();
    }

    /** helper for nice map comparison */
    private static String mapToString(Map m) {
        if (m == null) return "null";
        String res = "{\n";
        TreeSet sortedKeys = new TreeSet(m.keySet());
        boolean first = true;
        for (Object key : sortedKeys) {
            if (!first)
                res += ",\n";
            first = false;
            Object val = m.get(key);
            if (val== null) {
                res += "  \"" + key + "\" : null";
                continue;
            }
            if (val instanceof Number || val instanceof Boolean) {
                res += "  \"" + key + "\" : " + val;
                continue;
            }
            if (val instanceof String) {
                res += "  \"" + key + "\" : \"" + val+"\"";
                continue;
            }
            if (val instanceof Map) {
                res += "  \"" + key + "\" : " + mapToString((Map) val);
                continue;
            }
            if (val instanceof List) {
                res += "  \"" + key + "\" : " + listToString((List) val);
                continue;
            }
            throw new RuntimeException("Unsupported value type: "+val.getClass().getName());
        }
        return res+"\n}";
    }
    private static String listToString(List l) {
        if (l == null) return "null";
        String res = "[\n";
        boolean first = true;
        for (Object val : l) {
            if (!first)
                res += ", ";
            first = false;
            if (val== null) {
                res += "null";
                continue;
            }
            if (val instanceof Number || val instanceof Boolean) {
                res += val;
                continue;
            }
            if (val instanceof String) {
                res += "\"" + val+"\"";
                continue;
            }
            if (val instanceof Map) {
                res +=  mapToString((Map) val);
                continue;
            }
            if (val instanceof List) {
                res += listToString((List)val);
                continue;
            }
            throw new RuntimeException("Unsupported value type: "+val.getClass().getName());
        }
        return res + "\n]";

    }
}
